@@ -1,5 +1,5 @@
 /*
-** Copyright (C) 2002,2003 Erik de Castro Lopo <erikd@mega-nerd.com>
+** Copyright (C) 2002-2004 Erik de Castro Lopo <erikd@mega-nerd.com>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
@@ -16,23 +16,28 @@
 ** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
 
-#include "calc_snr.h"
-
 #include "config.h"
 
-#if (HAVE_LIBFFTW && HAVE_LIBRFFTW)
+#include "calc_snr.h"
+#include "util.h"
+
+#if (HAVE_FFTW3 == 1)
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <math.h>
 
-#include <rfftw.h>
+#include <fftw3.h>
 
 #define	MAX_SPEC_LEN	(1<<18)
 #define	MAX_PEAKS		10
 
+#define	ARRAY_LEN(x)	((int) (sizeof (x) / sizeof ((x) [0])))
+#define	MAX(x,y)		(((x) >= (y)) ? (x) : (y))
+
 static void log_mag_spectrum (double *input, int len, double *magnitude) ;
+static void smooth_mag_spectrum (double *magnitude, int len) ;
 static double find_snr (const double *magnitude, int len) ;
 
 typedef struct
@@ -49,64 +54,87 @@ calculate_snr (float *data, int len)
 	int k ;
 
 	if (len > MAX_SPEC_LEN)
-	{	printf (__FILE__ " : line %d : data length too large.\n", __LINE__) ;
+	{	printf ("%s : line %d : data length too large.\n", __FILE__, __LINE__) ;
 		exit (1) ;
 		} ;
 
-	/* Padd the data just a little to speed up the FFT. */
-	while ((len & 0x1F) && len < MAX_SPEC_LEN)
-	{	data [len] = 0.0 ;
-		len ++ ;
-		} ;
-
 	for (k = 0 ; k < len ; k++)
 		datacopy [k] = data [k] ;
 
+	/* Pad the data just a little to speed up the FFT. */
+	while ((len & 0x1F) && len < MAX_SPEC_LEN)
+	{	datacopy [len] = 0.0 ;
+		len ++ ;
+		} ;
+
 	log_mag_spectrum (datacopy, len, magnitude) ;
+	smooth_mag_spectrum (magnitude, len / 2) ;
 
 	snr = find_snr (magnitude, len) ;
 
 	return snr ;
 } /* calculate_snr */
 
+/*==============================================================================
+** There is a slight problem with trying to measure SNR with the method used
+** here; the side lobes of the windowed FFT can look like a noise/aliasing peak.
+** The solution is to smooth the magnitude spectrum by wiping out troughs
+** between adjacent peaks as done here.
+** This removes side lobe peaks without affecting noise/aliasing peaks.
+*/
+
+static void linear_smooth (double *mag, PEAK_DATA *larger, PEAK_DATA *smaller) ;
+
 static void
-log_mag_spectrum (double *input, int len, double *magnitude)
-{	rfftw_plan plan = NULL ;
+smooth_mag_spectrum (double *mag, int len)
+{	PEAK_DATA peaks [2] ;
 
-	double	maxval ;
-	int		k ;
+	int k ;
 
-	if (input == NULL || magnitude == NULL)
-		return ;
+	memset (peaks, 0, sizeof (peaks)) ;
 
-	plan = rfftw_create_plan (len, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE | FFTW_OUT_OF_PLACE) ;
-	if (plan == NULL)
-	{	printf (__FILE__ " : line %d : create plan failed.\n", __LINE__) ;
-		exit (1) ;
+	/* Find first peak. */
+	for (k = 1 ; k < len - 1 ; k++)
+	{	if (mag [k - 1] < mag [k] && mag [k] >= mag [k + 1])
+		{	peaks [0].peak = mag [k] ;
+			peaks [0].index = k ;
+			break ;
+			} ;
 		} ;
 
-	rfftw_one (plan, input, magnitude) ;
+	/* Find subsequent peaks ans smooth between peaks. */
+	for (k = peaks [0].index + 1 ; k < len - 1 ; k++)
+	{	if (mag [k - 1] < mag [k] && mag [k] >= mag [k + 1])
+		{	peaks [1].peak = mag [k] ;
+			peaks [1].index = k ;
 
-	/* (k < N/2 rounded up) */
-	maxval = 0.0 ;
-	for (k = 1 ; k < len / 2 ; k++)
-	{	magnitude [k] = sqrt (magnitude [k] * magnitude [k] + magnitude [len - k] * magnitude [len - k]) ;
-		maxval = (maxval < magnitude [k]) ? magnitude [k] : maxval ;
-		} ;
-	for ( ; k < len ; k++)
-		magnitude [k] = 0.0 ;
-
-	/* Don't care about DC component. Make it zero. */
-	magnitude [0] = 0.0 ;
-
-	/* log magnitude. */
-	for (k = 0 ; k < len ; k++)
-	{	magnitude [k] = magnitude [k] / maxval ;
-		magnitude [k] = (magnitude [k] < 1e-15) ? -200.0 : 20.0 * log10 (magnitude [k]) ;
+			if (peaks [1].peak > peaks [0].peak)
+				linear_smooth (mag, &peaks [1], &peaks [0]) ;
+			else
+				linear_smooth (mag, &peaks [0], &peaks [1]) ;
+			peaks [0] = peaks [1] ;
+			} ;
 		} ;
 
-	return ;
-} /* log_mag_spectrum */
+} /* smooth_mag_spectrum */
+
+static void
+linear_smooth (double *mag, PEAK_DATA *larger, PEAK_DATA *smaller)
+{	int k ;
+
+	if (smaller->index < larger->index)
+	{	for (k = smaller->index + 1 ; k < larger->index ; k++)
+			mag [k] = (mag [k] < mag [k - 1]) ? 0.999 * mag [k - 1] : mag [k] ;
+		}
+	else
+	{	for (k = smaller->index - 1 ; k >= larger->index ; k--)
+			mag [k] = (mag [k] < mag [k + 1]) ? 0.999 * mag [k + 1] : mag [k] ;
+		} ;
+
+} /* linear_smooth */
+
+/*==============================================================================
+*/
 
 static int
 peak_compare (const void *vp1, const void *vp2)
@@ -129,7 +157,7 @@ find_snr (const double *magnitude, int len)
 
 	/* Find the MAX_PEAKS largest peaks. */
 	for (k = 1 ; k < len - 1 ; k++)
-	{	if (magnitude [k-1] < magnitude [k] && magnitude [k] >= magnitude [k+1])
+	{	if (magnitude [k - 1] < magnitude [k] && magnitude [k] >= magnitude [k + 1])
 		{	if (peak_count < MAX_PEAKS)
 			{	peaks [peak_count].peak = magnitude [k] ;
 				peaks [peak_count].index = k ;
@@ -144,7 +172,7 @@ find_snr (const double *magnitude, int len)
 		} ;
 
 	if (peak_count < MAX_PEAKS / 2)
-	{	printf ("\n" __FILE__ " : line %d : bad peak_count (%d).\n\n", __LINE__, peak_count) ;
+	{	printf ("\n%s : line %d : bad peak_count (%d).\n\n", __FILE__, __LINE__, peak_count) ;
 		return -1.0 ;
 		} ;
 
@@ -159,6 +187,47 @@ find_snr (const double *magnitude, int len)
 	return snr ;
 } /* find_snr */
 
+static void
+log_mag_spectrum (double *input, int len, double *magnitude)
+{	fftw_plan plan = NULL ;
+
+	double	maxval ;
+	int		k ;
+
+	if (input == NULL || magnitude == NULL)
+		return ;
+
+	plan = fftw_plan_r2r_1d (len, input, magnitude, FFTW_R2HC, FFTW_ESTIMATE | FFTW_PRESERVE_INPUT) ;
+	if (plan == NULL)
+	{	printf ("%s : line %d : create plan failed.\n", __FILE__, __LINE__) ;
+		exit (1) ;
+		} ;
+
+	fftw_execute (plan) ;
+	
+	fftw_destroy_plan (plan) ;
+
+	/* (k < N/2 rounded up) */
+	maxval = 0.0 ;
+	for (k = 1 ; k < len / 2 ; k++)
+	{	magnitude [k] = sqrt (magnitude [k] * magnitude [k] + magnitude [len - k - 1] * magnitude [len - k - 1]) ;
+		maxval = (maxval < magnitude [k]) ? magnitude [k] : maxval ;
+		} ;
+
+	memset (magnitude + len / 2, 0, len / 2 * sizeof (magnitude [0])) ;
+
+	/* Don't care about DC component. Make it zero. */
+	magnitude [0] = 0.0 ;
+
+	/* log magnitude. */
+	for (k = 0 ; k < len ; k++)
+	{	magnitude [k] = magnitude [k] / maxval ;
+		magnitude [k] = (magnitude [k] < 1e-15) ? -200.0 : 20.0 * log10 (magnitude [k]) ;
+		} ;
+
+	return ;
+} /* log_mag_spectrum */
+
 #else /* ! (HAVE_LIBFFTW && HAVE_LIBRFFTW) */
 
 double
@@ -172,3 +241,12 @@ calculate_snr (float *data, int len)
 } /* calculate_snr */
 
 #endif
+
+/*
+** Do not edit or modify anything in this comment block.
+** The arch-tag line is a file identity tag for the GNU Arch 
+** revision control system.
+**
+** arch-tag: 7ae937c5-53a5-45d8-8aa2-793de9c35f0a
+*/
+
